from django.contrib import admin
from dynamic_preferences.models import GlobalPreferenceModel, UserPreferenceModel, SitePreferenceModel
from django import forms

class PreferenceChangeListForm(forms.ModelForm):
    """
    A form that integrate dynamic-preferences into django.contrib.admin
    """
    # Me must use an acutal model field, so we use raw_value. However, 
    # instance.value will be displayed in form.
    raw_value = forms.CharField()

    def __init__(self, *args, **kwargs):
        self.instance = kwargs.get('instance')
        super(PreferenceChangeListForm, self).__init__(*args, **kwargs)
        
        self.fields['raw_value'] = self.instance.preference.setup_field()            

    def save(self, *args, **kwargs):
        self.cleaned_data['raw_value'] = self.instance.preference.serializer.serialize(self.cleaned_data['raw_value'])
        return super(PreferenceChangeListForm, self).save(*args, **kwargs)

class GlobalPreferenceChangeListForm(PreferenceChangeListForm):
    class Meta:
        model = GlobalPreferenceModel

class UserPreferenceChangeListForm(PreferenceChangeListForm):
    class Meta:
        model = UserPreferenceModel

class DynamicPreferenceAdmin(admin.ModelAdmin):
    readonly_fields = ('name', 'section', 'value')
    fields = ("raw_value",)
    list_display = ('name', 'section', 'raw_value')
    list_editable = ('raw_value',)
    search_fields = ['name', 'section', 'raw_value']
    list_filter = ('section',)
    
    def get_changelist_form(self, request, **kwargs):
        return self.changelist_form


class GlobalPreferenceAdmin(DynamicPreferenceAdmin):
    form = GlobalPreferenceChangeListForm
    changelist_form = GlobalPreferenceChangeListForm
    
        
admin.site.register(GlobalPreferenceModel, GlobalPreferenceAdmin)


class UserPreferenceAdmin(DynamicPreferenceAdmin):
    form = UserPreferenceChangeListForm
    list_display = ('user',) + DynamicPreferenceAdmin.list_display
    search_fields = ['user__username'] + DynamicPreferenceAdmin.search_fields
    changelist_form = UserPreferenceChangeListForm
        
admin.site.register(UserPreferenceModel, UserPreferenceAdmin)



class SerializationError(Exception):
    pass

class BaseSerializer:
    """
        A serializer take a Python variable and returns a string that can be stored safely in database
    """
    exception = SerializationError

    @classmethod
    def serialize(cls, value, **kwargs):
        """
        Return a string from a Python var
        """
        raise NotImplementedError

    @classmethod
    def deserialize(cls, value, **kwargs):
        """
            Convert a python string to a var
        """
        raise NotImplementedError


class BooleanSerializer(BaseSerializer):

    true = (
        "True",
        "true",
        "TRUE",
        "1",
        "YES",
        "Yes"
        "yes",
    )

    false = (
        "False",
        "false",
        "FALSE",
        "0",
        "No",
        "no"
        "NO"
    )
    @classmethod
    def serialize(cls, value, **kwargs):
        """
            True is serialized to "1" to take less space
            same for False, with "0"
        """
        if value:
            return "1"
        else:
            return "0"

    @classmethod
    def deserialize(cls, value, **kwargs):

        if value in cls.true:
            return True

        elif value in cls.false:
            return False

        else:
            raise cls.exception("Value {0} can't be deserialized to a Boolean".format(value))


class IntSerializer(BaseSerializer):

    @classmethod
    def serialize(cls, value, **kwargs):
        if not isinstance(value, int):
            raise cls.exception('IntSerializer can only serialize int values')

        return value.__str__()

    @classmethod
    def deserialize(cls, value, **kwargs):
        try:
            return int(value)
        except:
            raise cls.exception("Value {0} cannot be converted to int")

from django.template import defaultfilters

class StringSerializer(BaseSerializer):

    @classmethod
    def serialize(cls, value, **kwargs):
        if not isinstance(value, basestring):
            raise cls.exception("Cannot serialize, value {0} is not a string".format(value))

        if kwargs.get("escape_html", False):
            return defaultfilters.force_escape(value)
        else:
            return value

    @classmethod
    def deserialize(cls, value, **kwargs):
        if not isinstance(value, basestring):
            raise cls.exception("Cannot deserialize, value {0} is not a string".format(value))

        return value


"""
Preferences are regular Python objects that can be declared within any django app.
Once declared and registered, they can be edited by admins (for :py:class:`SitePreference` and :py:class:`GlobalPreference`)
and regular Users (for :py:class:`UserPreference`)

UserPreference, SitePreference and GlobalPreference are mapped to corresponding PreferenceModel,
which store the actual values.

"""
from registries import user_preferences_registry, site_preferences_registry, global_preferences_registry
from dynamic_preferences.models import SitePreferenceModel, UserPreferenceModel, GlobalPreferenceModel


class BasePreference(object):
    """
    A base class that handle common logic  for preferences
    """

    #: The registry in which preference will be registered (:py:const:`registries.global_preferences`, :py:const:`registries.site_preferences` or :py:const:`registries.user_preferences`)
    registry = None

    #: The section under which the preference will be registered
    section = None

    #: The preference name
    name = ""

    #: A default value for the preference
    default = None

    #: The model corresponding to this preference type (:py:class:`SitePreference`, :py:class:`GlobalPreference` or :py:class:`UserPreference`)
    model = None

    def register(self):
        self.registry.register(name=self.name, section=self.section, preference=self)

    def to_model(self, **kwargs):
        """
        Retrieve a model instance corresponding to the Preference in database.
        This method will create the model instance if needed.


        :param kwargs: Keyword arguments that will be passed directly to queryset or new model
        :return: a :py:class:`models.BasePreferenceModel` instance
        """
        value = kwargs.pop('value', None)

        try:
            preference = self.model.objects.get(
                section=self.section,
                name=self.name,
                **kwargs
            )

        except self.model.DoesNotExist:
            preference = self.model(
                section=self.section,
                name=self.name,
                value=value,
                **kwargs
            )
            preference.save()

        return preference

    def identifier(self, separator="."):
        """
        Return the name and the section of the Preference joined with a separator, with the form `section<separator>name`
        """
        section = self.section or ""
        return unicode(separator.join([section, self.name]))

class GlobalPreference(BasePreference):
    """
        A preference that apply to a whole django installation
    """

    registry = global_preferences_registry
    model = GlobalPreferenceModel


class UserPreference(BasePreference):
    """
    Preference that is tied to a :py:class:`django.contrib.auth.models.User` instance
    """

    registry = user_preferences_registry
    model = UserPreferenceModel


class SitePreference(BasePreference):
    """
        Preference for each django.contrib.site
    """
    registry = site_preferences_registry
    model = SitePreferenceModel



from django.views.generic import TemplateView, FormView
from forms import preference_form_builder, user_preference_form_builder
from registries import user_preferences_registry

class PreferenceFormView(FormView):
    """Display a form for updating preferences of the given section provided via URL arg.
    If no section is provided, will dispaly a form for all fields of a given registry.
    """

    #: the registry for preference lookups
    registry = None

    #: will be used by :py:func:`forms.preference_form_builder` to create the form
    form_class = None

    template_name = "dynamic_preferences/form.html"

    def get_form_class(self, *args, **kwargs):
        section = self.kwargs.get('section', None)
        form_class = preference_form_builder(self.form_class, section=section)
        return form_class

    def get_context_data(self, *args, **kwargs):

        context = super(PreferenceFormView, self).get_context_data(*args, **kwargs)

        context['registry'] = self.registry

        return context

    def get_success_url(self):
        return self.request.path

    def form_valid(self, form):
        
        form.update_preferences()
        return super(PreferenceFormView, self).form_valid(form)

class UserPreferenceFormView(PreferenceFormView):
    """
    Will pass `request.user` to form_builder
    """
    registry = user_preferences_registry
    
    def get_form_class(self, *args, **kwargs):
        section = self.kwargs.get('section', None)
        form_class = user_preference_form_builder(user=self.request.user, section=section)
        return form_class


import os
from setuptools import setup, find_packages
import dynamic_preferences
README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-dynamic-preferences',
    version=dynamic_preferences.__version__,
    packages=find_packages(),
    include_package_data=True,
    license='GPLv3',  # example license
    description='A django app for registering dynamic global, site and user preferences',
    long_description=README,
    url='http://code.eliotberriot.com/eliotberriot/django-dynamic-preferences',
    author='Eliot Berriot',
    author_email='contact@eliotberriot.com',
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)', # example license
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        # Replace these appropriately if you are stuck on Python 2.
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)


from django import forms
from registries import global_preferences_registry, user_preferences_registry, site_preferences_registry

def preference_form_builder(form_base_class, preferences=[], **kwargs):
    """
    Return a form class for updating preferences
    :param form_base_class: a Form class used as the base. Must have a ``registry` attribute
    :param preferences: a list of :py:class:
    :param section: a section where the form builder will load preferences
    """
    registry = form_base_class.registry
    preferences_obj = []
    if len(preferences) > 0:
        # Preferences have been selected explicitly 
        for pref in preferences:
            if type(pref) == str:
                preferences_obj.append(registry.get(name=pref))
            elif type(pref) == tuple:
                preferences_obj.append(registry.get(name=pref[0], section=pref[1]))
            else:
                raise NotImplementedError("The data you provide can't be converted to a Preference object")
    else:
        # Try to use section param
        preferences_obj = registry.preferences(section=kwargs.get('section', None))

    fields = {}
    instances = []
    for preference in preferences_obj:
        f = preference.field
        model_kwargs = kwargs.get('model', {})
        instance = preference.to_model(**model_kwargs)
        f.initial = instance.value
        fields[preference.identifier()] = f
        instances.append(instance)

    form_class = type('Custom'+form_base_class.__name__, (form_base_class,), {})
    form_class.base_fields = fields
    form_class.preferences = preferences_obj
    form_class.instances = instances
    return form_class

def global_preference_form_builder(preferences=[], **kwargs):
    """
    A shortcut :py:func:`preference_form_builder(GlobalPreferenceForm, preferences, **kwargs)`
    """
    return preference_form_builder(GlobalPreferenceForm, preferences, **kwargs)

def user_preference_form_builder(user, preferences=[], **kwargs):
    """
    A shortcut :py:func:`preference_form_builder(UserPreferenceForm, preferences, **kwargs)`
    :param user: a :py:class:`django.contrib.auth.models.User` instance
    """
    return preference_form_builder(UserPreferenceForm, preferences, model={'user': user}, **kwargs)

def site_preference_form_builder(preferences=[], **kwargs):
    """
    A shortcut :py:func:`preference_form_builder(SitePreferenceForm, preferences, **kwargs)`
    """
    return preference_form_builder(SitePreferenceForm, preferences, **kwargs)

class PreferenceForm(forms.Form):

    registry = None

    def update_preferences(self, **kwargs):
        for instance in self.instances:
            instance.value = self.cleaned_data[instance.preference.identifier()]
            instance.save()

class GlobalPreferenceForm(PreferenceForm):

    registry = global_preferences_registry

class UserPreferenceForm(PreferenceForm):

    registry = user_preferences_registry

class SitePreferenceForm(PreferenceForm):

    registry = site_preferences_registry


"""
Preference models, queryset and managers that handle the logic for persisting preferences.
"""

from django.db import models
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.db.models.query import QuerySet
from utils import update
from django.conf import settings
from django.utils.functional import cached_property
from dynamic_preferences.registries import user_preferences_registry, site_preferences_registry, global_preferences_registry

class PreferenceSite(Site):

    class Meta:
        proxy = True
        app_label = 'dynamic_preferences'


class PreferenceUser(User):
    class Meta:
        proxy = True
        app_label = 'dynamic_preferences'


class BasePreferenceModel(models.Model):
    """
    A base model with common logic for all preferences models.
    """

    #: The section under which the preference is declared
    section = models.TextField(max_length=255, db_index=True, blank=True, null=True, default=None)

    #: a name for the preference
    name = models.TextField(max_length=255, db_index=True)

    #: a value, serialized to a string. This field should not be accessed directly, use :py:attr:`BasePreferenceModel.value` instead
    raw_value = models.TextField(null=True, blank=True)

    #: Keep a reference to the whole preference registry.
    #: In order to map the Preference Model Instance to the Preference object.
    registry = None

    class Meta:
        abstract = True
        app_label = 'dynamic_preferences'


    def __init__(self, *args, **kwargs):
        # Check if the model is already saved in DB

        v = kwargs.pop("value", None)

        super(BasePreferenceModel, self).__init__(*args, **kwargs)

        new = self.pk is None

        if new:
            if v is not None:
                self.value = v
            else:
                self.value = self.preference.default



    @cached_property
    def preference(self):
        return self.registry.get(section=self.section, name=self.name)

    def set_value(self, value):
        """
            Save serialized self.value to self.raw_value
        """
        self.raw_value = self.preference.serializer.serialize(value)

    def get_value(self):
        """
            Return deserialized self.raw_value
        """
        return self.preference.serializer.deserialize(self.raw_value)

    value = property(get_value, set_value)

    def __unicode__(self):
        return self.preference.identifier()

class GlobalPreferenceModel(BasePreferenceModel):

    registry = global_preferences_registry

    class Meta:
        unique_together = ('section', 'name')
        app_label = 'dynamic_preferences'

        verbose_name = "global preference"
        verbose_name_plural = "global preferences"


class UserPreferenceModel(BasePreferenceModel):

    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="preferences")
    registry = user_preferences_registry

    class Meta:
        unique_together = ('user', 'section', 'name')
        app_label = 'dynamic_preferences'
        verbose_name = "user preference"
        verbose_name_plural = "user preferences"

class SitePreferenceModel(BasePreferenceModel):

    site = models.ForeignKey(Site, related_name="preferences")
    registry = site_preferences_registry

    class Meta:
        unique_together = ('site', 'section', 'name')
        app_label = 'dynamic_preferences'
        verbose_name = "site preference"
        verbose_name_plural = "site preferences"


global_preferences = GlobalPreferenceModel.objects
site_preferences = SitePreferenceModel.objects
user_preferences = UserPreferenceModel.objects